#ifndef jctable_HPP
#define jctable_HPP

/**********************************************************

  Jctable.hpp
  v1.0
                manual page

  Language:     ANSIC++
  Platform:     1. Gateway 386/25
                   MS-DOS 6.2, win 95.4000.950
			    2. SP2
				   Unix, AIX

  Application:  AdJava
  Author:       Xinying Li, CIS, Syracuse University
                (315)443-4889, xli@npac.syr.edu


 ***********************************************************/



/*
    jctable class
    =============
    jctable is a container administing Java-CPP object table.

    constructors
    ============
    jctable c;                      void constructor
    jctable c2 = c1;                copy constructor

    operations
    ==========
    c2 = c1                    assignment
    len()                      return length of table
    size()                     return the physical size of the whole table,
						       include the empty items.
    long NewHandle()                append a new handle to the table.
    long Findhandle()
    IncRefCntr()               increase reference counter
    DecRefCntr()               decrease reference counter
   
    fields
    ======
    ObjPtr                     handle
    ObjRef                     reference counter
*/



/**********************************************************

               maintenance page

  Build Process:

  DOS:
  Files required :
		jctable.cpp jctable.hpp

  integrated environment:
  - close any open project
  - set Options\Compiler\Code generated\defines to
      TEST_JCTABLE
  - open jctable.cpp in edit window and run

  - command line build:
      bcc -w+ -DTEST_JCTABLE jctable.cpp

  UNIX: AIX
  Files required :
        jctable.C jctable.hpp

  integrated environment:
  - set Options\Compiler\Code generated\defines to
      TEST_JCTABLE
  - command line build:
      xlC -DTEST_JCTABLE jctable.C


 ***********************************************************/

/*

   Maintenance history:
   ====================
   - v1.0 : 10/28/1997
     . created jctable module



*/


/***********************************************************

               declarations

 ***********************************************************/


#include <iostream.h>
#include <stdlib.h>

template<class T> class jctable{

  public:
	jctable(int size=128);
//	jctable();
	jctable(const jctable<T> &t);
	~jctable();

	jctable<T>& operator=(const jctable<T> &t);
	long len() const;
	long size() const;

	T **ObjPtr; 
	long *ObjRef;

	void IncRefCntr(long);
	int DecRefCntr(long);
	long FindHandle(T *);
	long NewHandle(T *);
//	T *ResolvePtr();

  private:
    void remove(long);
	void resize(int);
	long _next;
	long _size;
	void _alloc(int size);
	void _alloc(const jctable<T> &t);
};

// return table length
template<class T>
inline long jctable<T>::len() const { return _next;}

template<class T>
inline long jctable<T>::size() const { return _size;}

// destructor
template <class T>
inline jctable<T>::~jctable() {
	if(ObjPtr)
	 delete[] ObjPtr;
	if(ObjRef)
	 delete[] ObjRef;
	_size = 0;
	_next = 0;
}

// allocate memory

template <class T>
inline void jctable<T>:: _alloc(int size) {
   _next = 0;
   ObjPtr = new T*[_size=size];
   if (!ObjPtr)
   { cout << "memory alloc failure in alloc" << endl;
	 exit(1);
	 }
   ObjRef = new long[size];
   if (!ObjRef)
   { cout << "memory alloc failure in alloc" << endl;
	 exit(1);
	 }
   for(int i=0;i<size;i++)
   {  ObjPtr[i]=NULL;
	  ObjRef[i]=0;
	  }
}

// allocate

template <class T>
inline void jctable<T>:: _alloc(const jctable<T> &t) {
	_alloc(t.size());
	for(int i=0;i<t.len();i++)
	{	 ObjPtr[i] = t.ObjPtr[i];
		 ObjRef[i] = t.ObjRef[i];
		 _next++;
		 }
}

// constructor

template <class T>
inline jctable<T> :: jctable(int size) {
	_alloc(size);
}

/*
template <class T>
inline jctable<T> :: jctable() {
    _alloc(128);
}
*/

// copy constructor

template <class T>
inline jctable<T> :: jctable(const jctable<T> &t) {
	_alloc(t);
}

// assignment

template <class T>
inline jctable<T>& jctable<T> :: operator=(const jctable<T> &t) {
	if(this == &t)
	   return *this;
	delete [] ObjPtr;
	delete [] ObjRef;
	_alloc(t);
	return *this;
}

//  resize

template <class T>
inline void jctable<T> :: resize(int newsize) {
	jctable<T> temp = *this;
	_alloc(newsize);
	for(int i=0;i<temp.len();i++)
	   {
		 ObjPtr[i] = temp.ObjPtr[i];
		 ObjRef[i] = temp.ObjRef[i];
		 _next++;
		 }
}

//  add new handle, resize by double.
template <class T>
inline long jctable<T>:: NewHandle(T *newObj) {
	long pos=0;
    while( pos <_size && ObjPtr[pos]!=NULL)
		 pos++;
    if(pos >= _size) resize(_size*2);
    ObjPtr[pos] = newObj;
    IncRefCntr(pos);
	if(pos==_next)
    	_next++;
    //printf("Template: _p %d _n %d\n",pos,_next);
    return pos;
}

/*
template <class T>
inline long jctable<T>:: NewHandle(T *newObj) {
	if(_next >= _size) resize(_size+10);
	ObjPtr[_next] = newObj;
	IncRefCntr(_next);
	_next++;
	return _next-1;
}
*/

// incRefCntr

template <class T>
inline void jctable<T>:: IncRefCntr(long TheCPPObj) {
	ObjRef[TheCPPObj]++;
}

// decRefCntr, return reference counter

template <class T>
inline int jctable<T>:: DecRefCntr(long TheCPPObj) {
	ObjRef[TheCPPObj]--;
//	if(ObjRef[TheCPPObj]==0)
//		remove(TheCPPObj);
//	   ObjPtr[TheCPPObj]=NULL;
	return ObjRef[TheCPPObj];
}

// remove an element from table

template <class T>
inline void jctable<T>:: remove(long n) {
	if(_next==0 || n<0 || _next<=n) return;
    for(int i=n;i<_next-1;i++)
	{   ObjPtr[i]=ObjPtr[i+1];
		ObjRef[i]=ObjRef[i+1];
	    }
	_next--;
}

// find Handle

template <class T>
inline long jctable<T>:: FindHandle(T *Obj) {
	for(int i=0;i<_next;i++)
		if(Obj == ObjPtr[i])
			return i;
	return _next;
}

/*
// ResolvePtr
template <class T>
inline T* jctable<T>:: ResolvePtr() {

}
*/

#endif
